# This script was written for Python 3.8.10. Compatibility with other Python versions may vary.
#
# This script is to convert Voxelmap waypoint files to Xaero's Minimap waypoint files.
#
# This script requires Xaero's Minimap to be installed in Minecraft for the waypoints to work.
#         https://www.curseforge.com/minecraft/mc-mods/xaeros-minimap
# This script also requires Xaero's Worldmap to be installed for multiworld to work
#         https://www.curseforge.com/minecraft/mc-mods/xaeros-world-map
#
# If run with no command line arguments,
#   the script will prompt the user for all the information it needs
# Run the script with the -h option to see the command line options
#
# I do not guarantee the effectiveness of this script. It has work for everything I have tried
#   with it so far, but there may be other cases where it fails. Use at your own risk,
#   and make backups before you start.
# If you find a bug, please report it to issues@orangeben.dev
#
# This software is licensed under GNU GPL V3.
# A copy of the license can be obtained at https://www.gnu.org/licenses/gpl-3.0.en.html

import os
import math
from datetime import datetime
import tkinter as tk
from tkinter import filedialog
from pathlib import Path
import sys

# Wether to print debug messages. Set by -d
debug = False
# Version of the script to be printed in exported files. This appears to be removed by Xaero's Minimap when it loads the file, but whatever.
scriptversion = "1.0"
# Allows you to prepend a filepath to the destination folders. Leave empty to work in the current directory
saveRoot = ''
# Command line arguments to be easily checked later
args = []
# The Minecraft color codes to convert Voxelmap colors to their closest Minecraft color
colors = (
    (0x00, 0x00, 0x00),
    (0x00, 0x00, 0xAA),
    (0x00, 0xAA, 0x00),
    (0x00, 0xAA, 0xAA),
    (0xAA, 0x00, 0x00),
    (0xAA, 0x00, 0xAA),
    (0xFF, 0xAA, 0x00),
    (0xAA, 0xAA, 0xAA),
    (0x55, 0x55, 0x55),
    (0x55, 0x55, 0xFF),
    (0x55, 0xFF, 0x55),
    (0x55, 0xFF, 0xFF),
    (0xFF, 0x55, 0x55),
    (0xFF, 0x55, 0xFF),
    (0xFF, 0xFF, 0x55),
    (0xFF, 0xFF, 0xFF)
)

def parseVoxelpoint(line):
    ''' Parses a Voxelmap waypoint, and returns a dictionary with the waypoint '''
    if(line.startswith("name:")):
        vwp = {}
        parts = line.split(',')
        vwp['line'] = line
        for part in parts:
            kv = part.split(':')
            vwp[kv[0]] = kv[1]
        return vwp
    else:
        return None

# Wether or not to print the debug messages for the colors. No command line option to enable
debugColor = False
# Xaero's waypoint format
# waypoint:name:initials:x:y:z:color:disabled:type:set:rotate_on_tp:tp_yaw:global
def convertToXaero(vwp):
    ''' Converts a Voxelmap waypoint dictionary to a Xaero's waypoint. Returns as a string'''
    name = vwp['name'] # the name of the wp
    initials = name[0].upper() # the initials of the wp to show on the icon
    x = vwp['x'] # the x position in the world
    y = vwp['y'] # the y position in the world
    z = vwp['z'] # The z position in the world
    color = 0 # This will be a bit trickey to figure out
    disabled = 'false' if vwp['enabled'] == 'true' else 'true'# if the waypoint is disabled
    type = '0' #Not sure what this does
    set = 'gui.xaero_default'# The default set
    rotate_on_tp = 'false' # Don't need since don't tp
    tp_yaw = '0' # Don't need since don't tp
    glbl = 'false' # Draw if too far

    # Okay, let's sort out the colors
    # First I'll put the colors in the range 0-255
    r = int(float(vwp['red']) * 255)
    g = int(float(vwp['green']) * 255)
    b = int(float(vwp['blue']) * 255)

    distance = 999 # Big number so every real value will be smaller
    # Now I'll calculate the total distance from the Voxelmap color to each Minecraft color,
    #   and remember which one is smallest.
    if debugColor:
        print(f"i:\t  r :   g :   b\t\t cr :  cg :  cb\t\td\tdst\tcol")
    for i in range(len(colors)):
        c = colors[i]
        dr = abs(r - c[0])
        dg = abs(g - c[1])
        db = abs(b - c[2])
        d = dr + dg + db
        if debugColor:
            print(f"{i}:\t{r: >3d} : {g: >3d} : {b: >3d}\t\t{c[0]: >3d} : {c[1]: >3d} : {c[2]: >3d}\t\t{d: 3d}\t{distance: 3d}\t{color: 1d}")
        if d < distance:
            color = i
            distance = d
    if debugColor:
        print(f"Color: {distance: 3d}\t{color: 1d}")

    # Make a string from the waypoint, and return it
    return f"waypoint:{name}:{initials}:{x}:{y}:{z}:{color}:{disabled}:{type}:{set}:{rotate_on_tp}:{tp_yaw}:{glbl}"

def writeXWP(vpt, file):
    ''' Writes Xaero's waypoint to the file, optionally including the Voxelmap point as a comment '''
    xpt = convertToXaero(vpt)
    if debug:
        file.write("# ")
        file.write(vpt["line"])
        file.write("\n")
    file.write(xpt)
    file.write('\n')

def pause():
    ''' Pauses and waits for the user, unless the user has specified the -y option '''
    if not 'y' in args:
        try:
            input("Press Enter to continue or Ctl + C to exit")
        except KeyboardInterrupt:
            print()
            exit()

def writeToDimensionFile(vwp, files):
    ''' Writes a waypoint to the correct file(s)'''
    # vDims are overworld, the_nether, the_end
    if "overworld" in vpt["dimensions"]:
        writeXWP(vpt, files[0])
    if "the_nether" in vpt["dimensions"]:
        writeXWP(vpt, files[1])
    if "the_end" in vpt["dimensions"]:
        writeXWP(vpt, files[2])

# File header for the Xaero's waypoint files. This appears to be removed by Xaero's Minimap when it loads the file, but whatever.
xfileheader = f"# Converted to Xaero Waypoint format from Voxelmap\n#   by Ben's converting script V{scriptversion} on {datetime.now().strftime('%m/%d/%y %H:%M:%S')}\n# waypoint:name:initials:x:y:z:color:disabled:type:set:rotate_on_tp:tp_yaw:global\n#\n"
def writeHeaders(files):
    ''' Writes the headers to the waypoint files '''
    files[0].write("#\n# overworld\n")
    files[0].write(xfileheader)
    files[1].write("#\n# the_nether\n")
    files[1].write(xfileheader)
    files[2].write("#\n# the_end\n")
    files[2].write(xfileheader)

def makeFolders(path):
    ''' Makes the folders for the specified file if they don't already exist '''
    path = path.rsplit(os.path.sep,1)[0]
    if debug:
        print(f"Path: {path}")
    Path(path).mkdir(parents=True, exist_ok=True)

def makeWorldmapDir(gamename, worldname):
    ''' Makes the folders for the worldmap file if you are using a multiworld server '''
    for part in ('null', 'DIM-1', 'DIM1'):
        foldername =  f"{saveRoot}{os.path.sep}XaeroWorldMap{os.path.sep}Multiplayer_{gamename}{os.path.sep}{part}{os.path.sep}mw${worldname}/"
        makeFolders(foldername)

def openDimConfigs(gamename):
    ''' Opens the dimension configs if you are using a multiworld server '''
    cfgs = []
    for part in ('null', 'DIM-1', 'DIM1'):
        dimconfigfilename = f"{saveRoot}{os.path.sep}XaeroWorldMap{os.path.sep}Multiplayer_{gamename}{os.path.sep}{part}{os.path.sep}dimension_config.txt"
        makeFolders(dimconfigfilename)
        try:
            cfg = open(dimconfigfilename, 'w')
            cfgs.append(cfg)
        except Exception:
            print("File error!")
            exit()
    return cfgs

def writeDimConfig(cfgs, line):
    ''' Writes a line to all of the dimension configs since they seem to all need to be the same '''
    for cfg in cfgs:
        cfg.write(line)

def openFiles(gamename, worldname, prefix='', suffix=''):
    ''' Opens the waypoitn files for writing. gamename is foldername, worldname is the dimension in that world '''

    files = []
    for dim in ('dim%0', 'dim%-1', 'dim%1'):
        filename = f"{saveRoot}{os.path.sep}XaeroWaypoints{os.path.sep}{gamename}{os.path.sep}{dim}{os.path.sep}{prefix}{worldname}{suffix}.txt"

        makeFolders(filename)

        try:
            file = open(filename, 'w')
            files.append(file)
        except Exception:
            print("File error!")
            exit()
    return files

def closeFiles(files):
    ''' Close files '''
    for file in files:
        file.close()

def printMsg(msg, end='\n'):
    ''' Print a message if quiet is not enabled '''
    if not 'q' in args:
        print(msg, end=end)

def debugPrint(msg, end='\n'):
    ''' Prints debug message if debug is enabled '''
    if debug:
        print(msg, end=end)

if __name__ == "__main__":
    # Parse command line arguments
    for arg in sys.argv:
        if arg.startswith('-'):
            for c in arg[1:]:
                if not c in args:
                    args.append(c)

    # Load filename if it is in the arguments
    filename = ''
    for arg in sys.argv:
        if not arg.startswith('-'):
            filename = arg
            filename = os.path.expanduser(filename)

    # Print help if needed
    if 'h' in args:
        print("Usage:")
        print("  convertWaypoints.py [-hsmyqd] [filename] [-hsmyqd]")
        print("Command line args:")
        print("  -h  Help - Show this message and exit")
        print("  -s  Singleplayer - Select singleplayer mode")
        print("  -m  Multiplayer - Select multiplayer mode")
        print("  -y  Yes - Skip all pause prompts")
        print("  -q  Quiet - Don't print anything. Overidden by -h and -d. Implies -y, You must specify a filename.")
        print("  -d  Debug - Print debug messages")
        print("  Arguments can be included seperatly or combined")
        print("    (ex: '-s -y -q' or '-syq' are equivalent)")
        eixt()

    # Makes sure that the quiet option has the correct set of other options set
    if 'd' in args:
        debug = True
        if 'q' in args:
            args.remove('q')
    if 'q' in args:
        if not 'y' in args:
            args.append('y')
        if filename == '':
            # If the program can't speak, and no file has been specified, there is nothing to do
            exit()

    # Ask user to run the script in the right locations
    printMsg("Please run this script in the Minecraft profile folder that you want the waypoints to end up in. If you do not, you will have to move the XaeroWaypoints folder to your Minecraft profile folder.")
    pause()

    # Ask the user for a file
    if filename == '':
        root = tk.Tk()
        root.withdraw()

        filename = filedialog.askopenfilename(title="Select Voxelmap file ...", filetypes=[('VoxelMap Waypoint File', '.points')])
        if len(filename) == 0:
            printMsg("Conversion canceled")
            exit()

    # Load the world mode from the command line if present
    mode = 'q'
    if 's' in args:
        mode = 's'
    elif 'm' in args:
        mode = 'm'
    if not (mode == 's' or mode == 'm') and 'q' in args:
        mode = 's'

    # Ask the user the type of the file (singleplayer/mutiplayer)
    while not (mode == 's' or mode == 'm'):
        printMsg("Is this is world singleplayer, multiplayer?")
        mode = input("Select whichever button you use to access the world? (s/m)")
    # Warn the user about multiworld compatibility
    if mode == 'm':
        printMsg("Xaero Minimap does not seem to handel multiworld servers as well as Voxelmap. You will have to manually select your wolrd from the gear in the Map menu. This script assumes that you are using both Xaero's Minimap and Xaero's Worldmap")
        pause()

    # Start actually converting the file
    with open(filename, 'r') as vpts:
        vwpts = []
        worldnames = []
        worlds = {}

        # Read the entire voxelmap file into memory
        for vpt_line in vpts:
            vpt_line = vpt_line.rstrip()
            if mode == 'm' and vpt_line.startswith("subworlds:"):
                # Read the subworlds
                worldnames = vpt_line[10:].split(',')
                if worldnames[-1] == '':
                    worldnames = worldnames[:-1]
                debugPrint(f"Found subworlds! {worldnames}")
            else:
                # Try to parse the line
                vpt = parseVoxelpoint(vpt_line)
                if vpt != None:
                    vwpts.append(vpt)

        #Get the game/server name
        gamename = filename.split(os.path.sep)[-1][:-7].split('~')[0]

        # Set the default worldname for singleplayer worlds
        worldname = "waypoints"
        if mode == 'm':
            # Add the default worldname for points that don't specify a world
            worldnames.append('default')
            # Write the general server config file
            serverconfigfilename = f"{saveRoot}{os.path.sep}XaeroWorldMap{os.path.sep}Multiplayer_{gamename}{os.path.sep}server_config.txt"
            makeFolders(serverconfigfilename)
            with open(serverconfigfilename, 'w') as servercfg:
                servercfg.write("multiworldType:1\nignoreServerLevelId:false\nignoreHeightmaps:false")
            # Create the dimension configs
            dcfg = openDimConfigs(gamename)
            writeDimConfig(dcfg, f'confirmedMultiworld:mw${worldnames[0]}\n')
            # Write dimension configs
            for i in range(len(worldnames)):
                worldname = worldnames[i]
                files = openFiles(gamename, worldname, prefix='mw$', suffix=f'_{i}')
                writeHeaders(files)
                makeWorldmapDir(gamename, worldname)
                writeDimConfig(dcfg, f'MWName:mw${worldname}:{worldname}\n')
                worlds[worldname] = files
            closeFiles(dcfg)
        else: # If mode is singleplayer:
            files = openFiles(gamename, "waypoints")
            writeHeaders(files)
            worlds['waypoints'] = files

        # Actually write the waypoints
        for vpt in vwpts:
            if vpt != None:
                if mode == 'm':
                    worldname = vpt['world']
                writeToDimensionFile(vpt, worlds[worldname if worldname != '' else 'default'])

        # Save everything and cleanly close the files.
        for fileset in files:
            closeFiles(files)

        printMsg(f"{gamename} has been converted.")

# The directory structures needed for the waypotins:
#
# For single world need:
# profile
#   └─ XaeroWaypoints
#       └─ gamename
#           ├─ dim%0
#           │   └─ waypoints.txt
#           ├─ dim%-1
#           │   └─ waypoints.txt
#           └─ dim%1
#               └─ waypoints.txt
#
# For multiworld need:
# profile
#   ├─ XaeroWaypoints
#   │   └─ Multiplayer_ipaddress
#   │       ├─ dim%0
#   │       │   ├─ mw$somename_#.txt
#   │       │   └─ mw$othername_#.txt
#   │       ├─ dim%-1
#   │       │   ├─ mw$somename_#.txt
#   │       │   └─ mw$othername_#.txt
#   │       └─ dim%1
#   │           ├─ mw$somename_#.txt
#   │           └─ mw$othername_#.txt
#   └─ XaeroWorldMap
#       └─ Multiplayer_ipaddress
#           ├─ server_config.txt
#           ├─ null
#           │   ├─ dimension_config.txt
#           │   ├─ mw$somename
#           │   │   └─
#           │   └─ mw$othername
#           │       └─
#           ├─ DIM-1
#           │   ├─ dimension_config.txt
#           │   ├─ mw$somename
#           │   │   └─
#           │   └─ mw$othername
#           │       └─
#           └─ DIM1
#               ├─ dimension_config.txt
#               ├─ mw$somename
#               │   └─
#               └─ mw$othername
#                   └─
